package com.ibm.oti.vm;
/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2010  All Rights Reserved
 */

import sun.misc.Unsafe;
import java.lang.reflect.Field;
import java.util.Map;

public class ORBVMHelpers {

	static Unsafe createUnsafe() {
		try {
			Field f = Unsafe.class.getDeclaredField("theUnsafe"); //$NON-NLS-1$
			f.setAccessible(true);
			return (Unsafe) f.get(null);
		} catch (Throwable t) {
		}
		return null;
	}

	static Unsafe unsafe = createUnsafe();

	/**
	 * This native is a re-implementation of the JVM_LatestUserDefinedLoader.
	 *
	 * It walks the stack to find the first user defined class loader from the top of stack.
	 * The walk will terminate early and return null if it encounters a frame that has been
	 * "marked" using the LUDCLMarkFrame() native.
	 *
	 * Returns the ClassLoader object or null if no user-defined loader is found
	 */
	public static native ClassLoader LatestUserDefinedLoader();

	private static native boolean is32Bit();

	private static native int getNumBitsInReferenceField();

	private static native int getNumBytesInReferenceField();

	private static native int getNumBitsInDescriptionWord();

	private static native int getNumBytesInDescriptionWord();

	private static native int getNumBytesInJ9ObjectHeader();

	private static native int getJ9ClassFromClass32(Class c);

	private static native int getTotalInstanceSizeFromJ9Class32(int j9clazz);

	private static native int getInstanceDescriptionFromJ9Class32(int j9clazz);

	private static native int getDescriptionWordFromPtr32(int descriptorPtr);

	private static native long getJ9ClassFromClass64(Class c);

	private static native long getTotalInstanceSizeFromJ9Class64(long j9clazz);

	private static native long getInstanceDescriptionFromJ9Class64(long j9clazz);

	private static native long getDescriptionWordFromPtr64(long descriptorPtr);

	// TODO: JavaDoc required please
	public static int getNumSlotsInObject(Class currentClass) {
		int numSlotsInObject;
		int numBytesInReferenceField = getNumBytesInReferenceField();
		if (is32Bit()) {
			int j9clazz = getJ9ClassFromClass32(currentClass);
			// Get the instance size out of the J9Class
			int instanceSize = getTotalInstanceSizeFromJ9Class32(j9clazz);
			numSlotsInObject = instanceSize / numBytesInReferenceField;
		} else {
			long j9clazz = getJ9ClassFromClass64(currentClass);
			// Get the instance size out of the J9Class
			long instanceSize = getTotalInstanceSizeFromJ9Class64(j9clazz);
			numSlotsInObject = (int) instanceSize / numBytesInReferenceField;
			if (instanceSize > Integer.MAX_VALUE / numBytesInReferenceField)
				numSlotsInObject = -1;
		}

		return numSlotsInObject;
	}

	// TODO: JavaDoc required please
	@SuppressWarnings("deprecation")
	public static int getSlotIndex(Field field) {
		// Depends on the definition of slot and whether it's 32 or 64 bit
		int bytesInHeader = getNumBytesInJ9ObjectHeader();
		int numBytesInReferenceField = getNumBytesInReferenceField();
		int index = (unsafe.fieldOffset(field) - bytesInHeader) / numBytesInReferenceField;
		return index;
	}

	// TODO: JavaDoc required please
	@SuppressWarnings("deprecation")
	public static void vmDeepCopy(Object srcObj, Object dest,
			Class currentClass, ClassLoader targetCL, int[] destOffsets,
			Class[] declaredType, ORBBase orbi, Map map, Class itsType) throws Exception {
		int numSlotsInObject;
		int numBitsInWord = getNumBitsInDescriptionWord();
		// Depends on the definition of slot and whether it's 32 or 64 bit
		int bytesInHeader = getNumBytesInJ9ObjectHeader();
		int numBytesInReferenceField = getNumBytesInReferenceField();
		int j9clazz = 0;
		int instanceSize = 0;
		int descriptorPtr = 0;
		long j9clazz64 = 0;
		long instanceSize64 = 0;
		long descriptorPtr64 = 0;

		if (is32Bit()) {
			j9clazz = getJ9ClassFromClass32(currentClass);
			// Get the instance size out of the J9Class
			instanceSize = getTotalInstanceSizeFromJ9Class32(j9clazz);
			// Get the pointer to instance descriptor out of the J9Class.  Tells which fields are refs and prims
			descriptorPtr = getInstanceDescriptionFromJ9Class32(j9clazz);
			numSlotsInObject = instanceSize / numBytesInReferenceField;
		} else {
			j9clazz64 = getJ9ClassFromClass64(currentClass);
			// Get the instance size out of the J9Class
			instanceSize64 = getTotalInstanceSizeFromJ9Class64(j9clazz64);
			// Get the pointer to instance descriptor out of the J9Class.  Tells which fields are refs and prims
			descriptorPtr64 = getInstanceDescriptionFromJ9Class64(j9clazz64);
			numSlotsInObject = (int) instanceSize64 / numBytesInReferenceField;
		}

		int countSlots = 0; //slotsInHeader;
		int descriptorWord = 0;
		long descriptorWord64 = 0;
		int bitIndex = 0;
		if (isDescriptorPointerTagged(descriptorPtr, descriptorPtr64)) {
			bitIndex++;
			if (is32Bit())
				descriptorWord = descriptorPtr >> 1;
			else
				descriptorWord64 = descriptorPtr64 >> 1;
		} else {
			if (is32Bit())
				descriptorWord = getDescriptionWordFromPtr32(descriptorPtr);
			else
				descriptorWord64 = getDescriptionWordFromPtr64(descriptorPtr64);
		}

		int srcIndex = 0;
		while (numSlotsInObject > 0) {
			int destIndex = destOffsets[(int) srcIndex];
			Class type = declaredType[srcIndex];

			if (destIndex >= 0) {
				if (isDescriptorPointerTagged(descriptorWord, descriptorWord64)) {
					Object fieldValue = unsafe.getObject(srcObj, bytesInHeader + (countSlots
							* numBytesInReferenceField));
					if (fieldValue != null) {
						Object result = orbi.quickCopyIfPossible(fieldValue,
								type, targetCL, map, itsType);

						if (result != null) {
							unsafe.putObject(dest, bytesInHeader + (destIndex
									* numBytesInReferenceField), result);
							countSlots++;
							if (countSlots >= numSlotsInObject)
								break;

							if (bitIndex == (numBitsInWord - 1)) {
								if (is32Bit())
									descriptorPtr = descriptorPtr
											+ ORBVMHelpers
													.getNumBytesInDescriptionWord();
								else
									descriptorPtr64 = descriptorPtr64
											+ ORBVMHelpers
													.getNumBytesInDescriptionWord();

								bitIndex = 0;

								if (is32Bit())
									descriptorWord = getDescriptionWordFromPtr32(descriptorPtr);
								else
									descriptorWord64 = getDescriptionWordFromPtr64(descriptorPtr64);
							} else {
								if (is32Bit())
									descriptorWord = descriptorWord >> 1;
								else
									descriptorWord64 = descriptorWord64 >> 1;
								bitIndex++;
							}

							srcIndex++;
							continue;
						}

						fieldValue = orbi.deepCopyIfRequired(fieldValue, type,
								targetCL, map, type);

						unsafe.putObject(dest, bytesInHeader + (destIndex
								* numBytesInReferenceField), fieldValue);
					}
				} else {
					unsafe.putInt(dest, bytesInHeader + (destIndex * numBytesInReferenceField),
							unsafe.getInt(srcObj, bytesInHeader + (countSlots
									* numBytesInReferenceField)));
				}
			}

			countSlots++;
			if (countSlots >= numSlotsInObject)
				break;

			if (bitIndex == (numBitsInWord - 1)) {
				if (is32Bit())
					descriptorPtr = descriptorPtr
							+ getNumBytesInDescriptionWord();
				else
					descriptorPtr64 = descriptorPtr64
							+ getNumBytesInDescriptionWord();

				bitIndex = 0;

				if (is32Bit())
					descriptorWord = getDescriptionWordFromPtr32(descriptorPtr);
				else
					descriptorWord64 = getDescriptionWordFromPtr64(descriptorPtr64);
			} else {
				if (is32Bit())
					descriptorWord = descriptorWord >> 1;
				else
					descriptorWord64 = descriptorWord64 >> 1;

				bitIndex++;
			}

			srcIndex++;
		}
	}

	private static boolean isDescriptorPointerTagged(int descriptorPtr, long descriptorPtr64) {
		if (is32Bit()) {
			if ((descriptorPtr & 1) == 1) {
				return true;
			}
		} else {
			if ((descriptorPtr64 & 1) == 1) {
				return true;
			}
		}

		return false;
	}

}
