package com.ibm.tools.attach.javaSE;
/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2009, 2010  All Rights Reserved
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import com.ibm.tools.attach.spi.AttachProvider;

/**
 * JavaSE-specific methods
 *
 */
class VirtualMachineDescriptorImpl extends com.ibm.tools.attach.VirtualMachineDescriptor {

	@Override
	/**
	 * @note this compares two remote VMs, not the objects
	 */
	public boolean equals(Object comparand) {
		return super.equals(comparand);
	}

	@Override
	/**
	 * 	@note this obtains the hashcode of the remote VM, not of this object
	 */
	public int hashCode() {
		return super.hashCode();
	}

	private final String notifierId;
	private final String replyFile;
	private final String attachSyncFileValue;
	private final long processId;
	private final long uid;

	/**
	 * @param provider AttachProvider associated with this VM
	 * @param id VM identifier (for machine use)
	 * @param displayName VM name for human use
	 */
	public VirtualMachineDescriptorImpl(AttachProvider provider, String id,
			String displayName) {
		super(provider, id, displayName);
		notifierId = null;
		attachSyncFileValue = null;
		replyFile = null;
		processId = 0;
		uid = 0;
	}

	/**
	 * @param provider AttachProvider associated with this VM
	 * @param id VM identifier (for machine use)
	 */
	public VirtualMachineDescriptorImpl(AttachProvider provider, String id) {
		super(provider, id);
		notifierId = null;
		attachSyncFileValue = null;
		replyFile = null;
		processId = 0;
		uid = 0;
	}

	/**
	 * Create an object for a target VM (other than the caller)
	 * @param provider AttachProvider which handles the target VM
	 * @param advert Advertisement file
	 */
	public VirtualMachineDescriptorImpl(AttachProvider provider,
			Advertisement advert) {
		super(provider, advert.getVmId(), advert.getDisplayName());
		notifierId = advert.getNotifier();
		replyFile = advert.getReplyFile();
		attachSyncFileValue = advert.getNotificationSync();
		processId = advert.getProcessId();
		uid = advert.getUid();
	}

	/**
	 *
	 * @return Operating system process ID for the target VM.
	 */
	long getProcessId() {
		return processId;
	}

	/**
	 * Create a new VirtualMachineDescriptor with data from a file
	 * @param provider AttachProvider which creates this.
	 * @param advertFile file containing text of the advertisement
	 * @return VirtualMachineDescriptor with information from advertisement file.
	 */
	static VirtualMachineDescriptorImpl fromAdvertisement(AttachProvider provider, File advertFile) {
		Advertisement advert = null;
		try {
			FileInputStream propStream = new FileInputStream(advertFile);
			advert = Advertisement.readAdvertisementFile(propStream); /* throws IOException if file cannot be read */
			propStream.close();
		} catch (IOException e) {
			IPC.logMessage("could not read advertisement file ", advertFile.getAbsolutePath()); //$NON-NLS-1$
			return null;
		}
		VirtualMachineDescriptorImpl vmd = new VirtualMachineDescriptorImpl(provider, advert);
		return vmd;
	}

	/**
	 * @return name of semaphore file
	 */
	String getNotifierId() {
		return notifierId;
	}

	/**
	 * @return replyFile path to reply file
	 */
	String getReplyFile() {
		return replyFile;
	}

	/**
	 * Figure out the path to the target's sync file, inferring it when the advertisement file cannot be read.
	 * @return file path
	 */
	String getAttachSyncFileValue() {
		String asFile = attachSyncFileValue;
		if (null == asFile) {
			/* don't have permissions to the advertisement file.  Infer the filename from the VMID */
			asFile = TargetDirectory.getNotificationSyncFile(id()).getAbsolutePath();
		}
		return asFile;
	}

	public long getUid() {
		return uid;
	}

}
